from kork import ast
from kork.ast_printer import AstPrinter
from kork.chain import CodeChain
from kork.environment import Environment
from kork.examples import (
    AbstractExampleRetriever,
    SimpleExampleRetriever,
    c_,
    format_examples,
    r_,
)
from kork.exceptions import KorkException
from kork.interpreter import InterpreterResult, run_interpreter
from kork.retrieval import AbstractContextRetriever, SimpleContextRetriever

__all__ = [
    "CodeChain",
    "ast",
    "AstPrinter",
    "KorkException",
    "InterpreterResult",
    "run_interpreter",
    "AbstractContextRetriever",
    "SimpleContextRetriever",
    "Environment",
    "c_",
    "r_",
    "format_examples",
    "AbstractExampleRetriever",
    "SimpleExampleRetriever",
]



"""API to import foreign functions."""
import inspect
import sys
import types
import typing
from typing import Any, Callable, List, Mapping, Tuple, TypedDict

from kork import ast

PY_VERSION = (sys.version_info.major, sys.version_info.minor)


class FunctionInfo(TypedDict):
    """Information about a function."""

    name: str
    args: List[Tuple[str, Any, Any]]
    return_type: Any
    docstring: str


def _type_repr(obj: Any) -> str:
    """Return the repr() of an object, special-casing types (internal helper).

    If obj is a type, we return a shorter version than the default
    type.__repr__, based on the module and qualified name, which is
    typically enough to uniquely identify a type.  For everything
    else, we fall back on repr(obj).
    """
    if PY_VERSION >= (3, 9):
        # Ignoring type for passing mypy testing with Python 3.8
        if isinstance(obj, types.GenericAlias):  # type: ignore[attr-defined]
            return repr(obj)
    if isinstance(obj, type):
        if obj.__module__ == "builtins":
            return obj.__qualname__
        return f"{obj.__module__}.{obj.__qualname__}"
    if obj is ...:
        return "..."
    if isinstance(obj, types.FunctionType):
        return obj.__name__
    return repr(obj)


def _type_repr_wrapper(obj: Any) -> str:
    """A hack to remove all references to `typing.` in type strings.

    There's likely a much better solution, but this will do for prototyping.

    Args:
        obj (Any): The object to get the type string of.

    Returns:
        str: The type string of the object.
    """
    return _type_repr(obj).replace("typing.", "")


def _get_stringified_type(type_hints: Mapping[str, Any], arg_name: str) -> str:
    """Get string version of an argument type from a mapping of type hints.

    Args:
        type_hints (Mapping[str, Any]): Mapping of argument names to types.
        arg_name (str): Name of the argument or `return` for the return type.

    Returns:
        str: String representation of the type if it exists, otherwise Any.
    """
    if arg_name in type_hints:
        return _type_repr_wrapper(type_hints[arg_name])
    else:
        return "Any"


def _get_function_info(func: Callable) -> FunctionInfo:
    """Extract information about a function.

    Args:
        func (callable): The function to get information about.

    Returns:
        A dictionary with the following keys:
            - name: The name of the function.
            - args: A list of tuples containing the name and type of each argument.
            - return_type: The type of the value returned by the function.
            - docstring: The docstring of the function.
    """
    # Get the name of the function
    name = func.__name__

    try:
        # Get information about the function's arguments
        type_hints = typing.get_type_hints(func)
    except ValueError:
        type_hints = {}

    try:
        signature = inspect.signature(func)
        args = []
        for arg_name, param in signature.parameters.items():
            arg_type = _get_stringified_type(type_hints, arg_name)
            arg_default = (
                param.default if param.default != inspect.Parameter.empty else None
            )
            args.append((arg_name, arg_type, arg_default))

        # Get the return type of the function
        return_type = _get_stringified_type(type_hints, "return")
    except ValueError:  # TODO: Exception is too broad, limit to correct scope
        return_type = "Any"
        args = []

    # Get the docstring of the function
    docstring = func.__doc__

    # Construct and return a dictionary with the function information
    return {
        "name": name,
        "args": args,
        "return_type": return_type,
        "docstring": docstring.strip() if docstring else "",
    }


def _convert_to_kork_expr(value: Any) -> ast.Expr:
    """Convert mixed-expression value to a kork expression.

    This function converts a potentially mixed python, kork expression
    into a pure kork expression.

    Args:
        value (Any): The value to convert.

    Returns:
        An `kork` expression.
    """
    if isinstance(value, ast.Expr):
        # If it's already a kork expression, return as is
        return value
    elif isinstance(value, (int, float, bool, type(None), str)):
        return ast.Literal(value=value)
    elif isinstance(value, (list, tuple)):
        return ast.List_(elements=[_convert_to_kork_expr(v) for v in value])
    else:
        raise ValueError(f"Cannot convert {value} to an `kork` expression.")


# PUBLIC API


def to_extern_func_def(func: Callable) -> ast.ExternFunctionDef:
    """Convert a python function to a kork external function definition."""
    func_info = _get_function_info(func)
    params = [
        ast.Param(name=arg_name, type_=arg_type)
        for arg_name, arg_type, _ in func_info["args"]
    ]
    return ast.ExternFunctionDef(
        name=func_info["name"],
        params=ast.ParamList(params=params),
        return_type=func_info["return_type"],
        implementation=func,
        doc_string=func_info["docstring"],
    )


def to_kork_function_call(func: Callable, *args: Any) -> ast.FunctionCall:
    """Convert a python function call to a kork function call."""
    return ast.FunctionCall(
        name=func.__name__,
        args=[_convert_to_kork_expr(arg) for arg in args],
    )



"""The AST for the language.

The AST is a bit messy right now in terms of what's a statement vs. an expression,
and will likely need to be cleaned up a bit in the near future.
"""
from __future__ import annotations

import abc
import dataclasses
from typing import Any, Callable, Optional, Sequence, TypeVar, Union

T = TypeVar("T")


class Visitor(abc.ABC):
    """Abstract visitor interface."""


def _to_snake_case(name: str) -> str:
    """Convert a name into snake_case."""
    snake_case = ""
    for i, char in enumerate(name):
        if char.isupper() and i != 0:
            snake_case += "_" + char.lower()
        else:
            snake_case += char.lower()
    return snake_case


@dataclasses.dataclass(frozen=True)
class Expr(abc.ABC):
    """Abstract expression."""

    def accept(self, visitor: Visitor, **kwargs: Any) -> Any:
        """Accept implementation for a visitor."""
        return getattr(visitor, f"visit_{_to_snake_case(self.__class__.__name__)}")(
            self, **kwargs
        )


@dataclasses.dataclass(frozen=True)
class Literal(Expr):
    """A literal expression."""

    value: Union[float, int, bool, None, str]


@dataclasses.dataclass(frozen=True)
class Grouping(Expr):
    """A grouping expression."""

    expr: Expr


@dataclasses.dataclass(frozen=True)
class Unary(Expr):
    """A unary expression."""

    operator: str
    right: Expr


@dataclasses.dataclass(frozen=True)
class Binary(Expr):
    """A binary expression."""

    left: Expr
    operator: str
    right: Expr


@dataclasses.dataclass(frozen=True)
class Assign(Expr):
    """Assignment statement for a variable."""

    name: str
    value: Expr


@dataclasses.dataclass(frozen=True)
class Variable(Expr):
    """Variable reference."""

    name: str


@dataclasses.dataclass(frozen=True)
class List_(Expr):
    """List literal."""

    elements: Sequence[Expr]


@dataclasses.dataclass(frozen=True)
class Stmt(abc.ABC):
    """Abstract statement."""

    def accept(self, visitor: Visitor, **kwargs: Any) -> Any:
        """Accept implementation for a visitor."""
        return getattr(visitor, f"visit_{_to_snake_case(self.__class__.__name__)}")(
            self, **kwargs
        )


@dataclasses.dataclass(frozen=True)
class Param(Stmt):
    """Represent a function parameter."""

    name: str
    type_: str


@dataclasses.dataclass(frozen=True)
class ParamList(Stmt):
    """Represent a list of function parameters."""

    params: Sequence[Param]


@dataclasses.dataclass(frozen=True)
class FunctionDef(Stmt):
    """Represent a function definition with an implementation."""

    name: str
    params: ParamList
    body: Sequence[Union[Stmt, Expr]]
    # This is just an annotation so it's a string.
    return_type: str


@dataclasses.dataclass(frozen=True)
class ExternFunctionDef(Stmt):
    """External function definition."""

    name: str
    params: ParamList
    # This is just an annotation so it's a string.
    return_type: str
    implementation: Optional[Callable] = None
    doc_string: str = ""

    def add_implementation(self, implementation: Callable) -> "ExternFunctionDef":
        """Add an implementation to an external function definition."""
        if self.implementation is not None:
            raise ValueError("Cannot add implementation to an already implemented func")
        return ExternFunctionDef(
            name=self.name,
            params=self.params,
            return_type=self.return_type,
            implementation=implementation,
        )


@dataclasses.dataclass(frozen=True)
class FunctionCall(Expr):
    """Represent a function call."""

    name: str  # Name of function being invoked
    args: Sequence[Expr]  # Arguments to the function call


@dataclasses.dataclass(frozen=True)
class VarDecl(Stmt):
    """Represent a variable declaration."""

    name: str  # Name of variable
    value: Expr  # Expression to assign to the variable


@dataclasses.dataclass(frozen=True)
class Program(Stmt):
    """Represent a program."""

    stmts: Sequence[Union[Stmt, Expr]]



"""Definitions for custom Kork exceptions."""


class KorkException(Exception):
    """Generic Kork exception."""


class LLMParseException(KorkException):
    """Failed to parse LLM output."""


class KorkSyntaxException(KorkException):
    """Exceptions raised during syntax parsing."""


class KorkInterpreterException(KorkException):
    """Exceptions raised during interpretation."""


class KorkRunTimeException(KorkInterpreterException):
    """An exception that is raised during Kork interpreter runtime."""



from __future__ import annotations

import copy
import dataclasses
from dataclasses import field
from typing import Any, Dict, List, Mapping, Optional, Sequence

from kork import ast
from kork.exceptions import KorkRunTimeException


@dataclasses.dataclass
class Environment:
    """Environment for storing variables and function definitions."""

    parent: Optional[Environment] = None
    variables: Dict[str, Any] = field(default_factory=dict)

    def get_symbol(self, name: str) -> Any:
        """Get a symbol from the environment.

        Args:
            name: symbol name to lookup

        Returns:
            SymbolValue for the given symbol if it exists, otherwise raises an error
        """
        if name in self.variables:
            return self.variables[name]
        elif self.parent:
            return self.parent.get_symbol(name)
        else:
            raise KorkRunTimeException(f"Variable `{name}` not found")

    def set_symbol(self, name: str, value: Any) -> Any:
        """Set a symbol in the environment."""
        # TODO: We need to determine whether want the variable to be
        #       declared prior to be being set.
        #       At the moment, this allows for `x = 1` without having to declare the
        #       variable.
        self.variables[name] = value
        return value

    def clone(self) -> Environment:
        """Clone a root level environment, with deep copy on variables.

        Cloning functionality is provided because the environment is
        a mutable variable.

        Biggest danger is mutating the variables dict which the caller
        may retain a reference to.

        TODO(Eugene): Refactor to make this less dangerous.
        """
        if self.parent is not None:
            raise AssertionError("Cannot clone an environment with a parent.")
        return Environment(parent=None, variables=copy.deepcopy(self.variables))

    def list_external_functions(self) -> List[ast.ExternFunctionDef]:
        """Get a list of the externally defined foreign functions."""
        return [
            func
            for func in self.variables.values()
            if isinstance(func, ast.ExternFunctionDef)
        ]


def create_environment(
    extern_function_defs: Sequence[ast.ExternFunctionDef],
    external_variables: Optional[Mapping[str, Any]] = None,
) -> Environment:
    """Create a new environment with pre-populated state."""
    variables: Dict[str, Any] = {func.name: func for func in extern_function_defs}
    environment = Environment(parent=None, variables=variables)

    if external_variables:
        for name, value in external_variables.items():
            # At the moment, let's not clone the variable so that it can be
            # mutated if the goal is to do mutation.
            environment.set_symbol(name, value)
    return environment



"""A prompt adapter to allow working with both regular LLMs and Chat LLMs.

The prompt adapter supports breaking the prompt into:

1) Instruction Section
2) (Optional) Example Section
"""
from typing import Any, Callable, List, Sequence, Tuple

from langchain import BasePromptTemplate, PromptTemplate
from langchain.schema import BaseMessage, HumanMessage, PromptValue, SystemMessage
from pydantic import Extra


class FewShotPromptValue(PromptValue):
    """Integration with langchain prompt format."""

    string: Callable[[], str]
    messages: Callable[[], List[BaseMessage]]

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True

    def to_string(self) -> str:
        """Format the prompt to a string."""
        return self.string()

    def to_messages(self) -> List[BaseMessage]:
        """Get materialized messages."""
        return self.messages()


class FewShotTemplate(BasePromptTemplate):
    """Code prompt template."""

    instruction_template: PromptTemplate
    examples: Sequence[Tuple[str, str]] = ()

    @property
    def _prompt_type(self) -> str:
        """Get the prompt type."""
        return "FewShotTemplate"

    def format_prompt(self, **kwargs: Any) -> PromptValue:
        """Format the prompt."""
        if len(self.input_variables) != 1:
            raise AssertionError(
                f"Expected 1 input variable, got {len(self.input_variables)}"
            )

        query_key = self.input_variables[0]
        query = kwargs[query_key]

        def _lazy_string() -> str:
            """Lazy string."""
            return self.to_string(query)

        def _lazy_messages() -> List[BaseMessage]:
            """Lazy messages."""
            return self.to_messages(query)

        return FewShotPromptValue(
            string=_lazy_string,
            messages=_lazy_messages,
        )

    def format(self, **kwargs: Any) -> str:
        """Deprecated format method."""
        raise NotImplementedError()

    def to_string(self, query: str) -> str:
        """Format the prompt to a string."""
        instruction_section = self.instruction_template.format()

        examples_block = ""

        for input_example, output_example in self.examples:
            examples_block += f"Input: {input_example}\n\nOutput: {output_example}\n\n"

        return instruction_section + examples_block + f"Input: {query}\n\nOutput:"

    def to_messages(self, query: str) -> List[BaseMessage]:
        """Get materialized messages."""
        instruction_segment = self.instruction_template.format().strip()
        messages: List[BaseMessage] = [SystemMessage(content=instruction_segment)]

        for input_example, output_example in self.examples:
            messages.append(HumanMessage(content=input_example.strip()))
            messages.append(SystemMessage(content=output_example.strip()))

        messages.append(HumanMessage(content=query.strip()))

        return messages



"""Interface to specify kork examples easily."""
import abc
from typing import Any, Callable, List, Literal, Sequence, Tuple, Union

from kork import ast
from kork.ast_printer import AbstractAstPrinter
from kork.foreign_funcs import to_kork_function_call
from kork.utils import wrap_in_tag


def _add_result_variable(expr: ast.Expr) -> ast.Program:
    """Assembles a solution from a program."""
    stmt = ast.VarDecl(name="result", value=expr)
    return ast.Program(stmts=[stmt])


# PUBLIC API

# Use to denote different types of formatters for the input.
InputFormatter = Union[
    Literal["text_prefix"],
    Literal["triple_quotes"],
    Literal["markdown_text"],
    None,
    Callable[[str], str],
]


def format_text(text: str, input_formatter: InputFormatter = None) -> str:
    """An encoder for the input text.

    Args:
        text: the text to encode
        input_formatter: the formatter to use for the input
            * None: use for single sentences or single paragraphs, no formatting
            * triple_quotes: surround input with \"\"\", use for long text
            * text_prefix: same as triple_quote but with `TEXT: ` prefix
            * Callable: user provided function

    Returns:
        The encoded text if it was encoded
    """
    if input_formatter == "text_prefix":
        return 'Text: """\n' + text + '\n"""'
    elif input_formatter == "triple_quotes":
        return '"""\n' + text + '\n"""'
    elif input_formatter == "markdown_text":
        return "```text\n" + text + "\n```"
    elif input_formatter is None:
        return text
    else:
        raise NotImplementedError(
            f'No support for input encoding "{input_formatter}". '
            ' Use one of "long_text" or None.'
        )


# Alias to create a kork function call
def c_(name: Callable, *args: Any) -> ast.FunctionCall:
    """Create a kork function call."""
    return to_kork_function_call(name, *args)


# Alias to assign the expression to a variable called `result`
def r_(expr: ast.Expr) -> ast.Program:
    """Assign last program expression to a result variable."""
    return _add_result_variable(expr)


def format_examples(
    language_name: str,
    examples: Sequence[Tuple[str, ast.Program]],
    ast_printer: AbstractAstPrinter,
    input_formatter: InputFormatter = None,
) -> List[Tuple[str, str]]:
    """Format examples."""
    formatted_examples = []
    for example_input, desired_output in examples:
        formatted_input = format_text(example_input, input_formatter=input_formatter)
        formatted_examples.append(
            (
                formatted_input,
                wrap_in_tag("code", ast_printer.visit(desired_output)),
            )
        )

    return formatted_examples


class AbstractExampleRetriever(abc.ABC):
    """Abstract interface for an example retriever.

    An example interface must implement the `retrieve` method which
    returns a list of relevant examples based on the given query.
    """

    @abc.abstractmethod
    def retrieve(self, query: str) -> List[Tuple[str, str]]:
        """Retrieve examples."""


class SimpleExampleRetriever(AbstractExampleRetriever):
    """Simple example retriever.

    Simple example that returns the examples it was initialized with.

    Supports initialization from a list of programs.

    Example:

    .. code-block:: python

        from kork import SimpleExampleRetriever, AstPrinter, c_, r_

        simple_example_retriever = SimpleExampleRetriever.from_programs(
            language_name="kork",
            examples=[
                ("add 1 2", r_(c_(add, 1, 2))),
                ("add 1 2 3", r_(c_(add, 1, 2, 3))),
                ],
            ast_printer=AstPrinter(),
        )

        examples = simple_example_retriever.retrieve("add 1 2")
    """

    def __init__(
        self,
        examples: Sequence[Tuple[str, str]],
    ) -> None:
        """Initialize the retriever with a list of examples."""
        self._examples = examples

    def retrieve(self, query: str) -> List[Tuple[str, str]]:
        """Retrieve examples that best match the given query."""
        return list(self._examples)

    @classmethod
    def from_programs(
        cls,
        language_name: str,
        examples: Sequence[Tuple[str, ast.Program]],
        ast_printer: AbstractAstPrinter,
    ) -> "SimpleExampleRetriever":
        """Create a simple example retriever from programs.

        Args:
            language_name: The language name to use for the markdown code blocks.
            examples: A sequence of tuples of the form (query, desired program).
            ast_printer: The ast printer to use to format the desired output.

        Returns:
            A simple example retriever.
        """
        return cls(
            examples=format_examples(
                language_name=language_name,
                examples=examples,
                ast_printer=ast_printer,
            )
        )



import re
from typing import Optional


def wrap_in_tag(tag_name: str, content: str) -> str:
    """Wrap the content in an HTML style tag."""
    return f"<{tag_name}>{content}</{tag_name}>"


def unwrap_tag(tag_name: str, text: str) -> Optional[str]:
    """Extract content located inside a tag."""
    pattern = f"<{tag_name}>(.*?)</{tag_name}>"
    content = re.compile(pattern, re.DOTALL)
    match = content.search(text)
    if match:
        return match.group(1)
    else:
        return None


def unwrap_code(language_name: str, text: str) -> Optional[str]:
    """Extract code located inside a code tag."""
    pattern = f"```{language_name}\n(.*?)\n```"
    content = re.compile(pattern, re.DOTALL)
    match = content.search(text)
    if match:
        return match.group(1)
    else:
        return None



"""Implementation of a programming chain."""
from __future__ import annotations

from typing import (
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypedDict,
    Union,
    cast,
)

from langchain import LLMChain
from langchain.chains.base import Chain
from langchain.prompts import PromptTemplate

try:
    from langchain.base_language import BaseLanguageModel
except ImportError:
    from langchain.schema import BaseLanguageModel
from pydantic import Extra

from kork import ast
from kork.ast_printer import AbstractAstPrinter, AstPrinter
from kork.environment import Environment, create_environment
from kork.examples import (
    AbstractExampleRetriever,
    InputFormatter,
    SimpleExampleRetriever,
    format_examples,
    format_text,
)
from kork.exceptions import LLMParseException
from kork.interpreter import InterpreterResult, run_interpreter
from kork.prompt_adapter import FewShotTemplate
from kork.retrieval import AbstractContextRetriever, FuncLike, SimpleContextRetriever
from kork.utils import unwrap_tag

_INSTRUCTION_PROMPT = """\
You are programming in a language called "{language_name}".

You are an expert programmer and must follow the instructions below exactly.

Your goal is to translate a user query into a corresponding and valid {language_name}
program.

{external_functions_block}

Do not assume that any other functions except for the ones listed above exist.

Wrap the program in <code> and </code> tags.

Store the solution to the query in a variable called "result".

Here is a sample valid program:

<code>
var x = 1 # Assign 1 to the variable x
var result = 1 + 2 # Calculate the sum of 1 + 2 and assign to result
var result = x # Assign the value of x to result
</code>

Guidelines:
- Do not use operators, instead invoke appropriate external functions.
- Do not declare functions, do not use loops, do not use conditionals.
- Solve the problem only using variable declarations and function invocations.

Begin!
"""


# PUBLIC API


DEFAULT_INSTRUCTION_PROMPT = PromptTemplate(
    input_variables=[
        "language_name",
        "external_functions_block",
    ],
    template=_INSTRUCTION_PROMPT,
)


class CodeResult(TypedDict):
    """Result of a code chain."""

    errors: Sequence[Exception]
    raw: str
    code: str
    environment: Optional[Environment]  # Returned on success


# PUBLIC API

ExampleTuple = Tuple[str, ast.Program]


class CodeChain(Chain):
    """A coding chain."""

    llm: BaseLanguageModel
    """The language model to use for the coding chain."""
    interpreter: Callable[[str, Environment], InterpreterResult]
    """The interpreter to use for the coding chain."""
    ast_printer: AbstractAstPrinter
    """The AST printer to use for the coding chain."""
    context_retriever: Optional[AbstractContextRetriever] = None
    """Context to inject into the environment and prompt.
    
    At the moment, this context is limited to external functions.
    """
    example_retriever: Optional[AbstractExampleRetriever] = None
    """Examples that should be added to the prompt."""
    instruction_template: PromptTemplate = DEFAULT_INSTRUCTION_PROMPT
    """Template for the instruction prompt."""
    input_key: str = "query"
    # Smirking kat language
    language_name: str = "😼"
    input_formatter: InputFormatter = "triple_quotes"

    """The smirking kat programming language; aka Kork; aka 😼"""

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.allow
        arbitrary_types_allowed = True

    @property
    def input_keys(self) -> List[str]:
        """Will be whatever keys the prompt expects.

        :meta private:
        """
        return [self.input_key]

    @property
    def output_keys(self) -> List[str]:
        """Output keys to return.

        :meta private:
        """
        return ["raw", "errors", "code", "environment"]

    def prepare_context(
        self, query: str, variables: Optional[Mapping[str, Any]] = None
    ) -> Tuple[Environment, FewShotTemplate]:
        """Get the pre-populated environment and the few shot template.

        Args:
            query: The query to prepare the context for.
            variables: Any variables that should be added to the context.

        Returns:
            The prepopulated environment and a pre-formatted few shot template.
        """
        if self.context_retriever:
            external_function_definitions = self.context_retriever.retrieve(query)
        else:
            external_function_definitions = []

        environment = create_environment(external_function_definitions, variables)

        if self.example_retriever:
            formatted_examples = self.example_retriever.retrieve(query)
        else:
            formatted_examples = []

        external_functions = "\n".join(
            [
                self.ast_printer.visit(func_declaration)
                for func_declaration in environment.list_external_functions()
            ]
        )

        input_variables = self.instruction_template.input_variables

        formatting_variables = {}

        if "language_name" in input_variables:
            formatting_variables["language_name"] = self.language_name

        if external_functions:
            external_functions_block = (
                "You have access to the following external functions:\n\n"
                f"```{self.language_name}\n"
                f"{external_functions}\n"
                "```\n"
            )
        else:
            external_functions_block = "Do not assume any functions exist.\n"

        if "external_functions_block" in input_variables:
            formatting_variables["external_functions_block"] = external_functions_block

        prompt = self.instruction_template.partial(
            **formatting_variables,
        )

        return environment, FewShotTemplate(
            instruction_template=prompt,
            examples=formatted_examples,
            input_variables=["query"],
        )

    def _call(self, inputs: Dict[str, str]) -> CodeResult:  # type: ignore
        """Call the chain."""
        # Note not using original query!!
        # We remove the leading and trailing whitespace from the query
        # to make things a bit more robust.
        query = inputs[self.input_key].strip()

        variables: Mapping[str, Any] = cast(
            Mapping[str, Any], inputs.get("variables", {})
        )

        if not isinstance(variables, dict):
            raise TypeError(
                f"Variables must be a dictionary, got {type(variables)} instead."
            )

        environment, few_shot_template = self.prepare_context(query, variables)

        chain = LLMChain(
            prompt=few_shot_template,
            llm=self.llm,
        )

        formatted_query = format_text(query, self.input_formatter)
        llm_output = cast(str, chain.predict_and_parse(query=formatted_query))
        code = unwrap_tag("code", llm_output)

        if not code:
            return {
                "errors": [
                    LLMParseException(
                        "Could not find code block. Please make sure the code "
                        "is wrapped in a code block."
                    )
                ],
                "raw": llm_output,
                "code": "",
                "environment": None,
            }

        interpreter_result = self.interpreter(code, environment)

        if interpreter_result["errors"]:
            return {
                "errors": interpreter_result["errors"],
                "raw": llm_output,
                "code": code,
                "environment": None,
            }

        return {
            "errors": [],
            "raw": llm_output,
            "code": code,
            "environment": interpreter_result["environment"],
        }

    @property
    def _chain_type(self) -> str:
        """Used for serialization."""
        return "code_chain"

    @classmethod
    def from_defaults(
        cls,
        *,
        llm: BaseLanguageModel,
        interpreter: Callable[[str, Environment], InterpreterResult] = run_interpreter,
        ast_printer: AbstractAstPrinter = AstPrinter(),
        context: Union[AbstractContextRetriever, Sequence[FuncLike], None] = None,
        examples: Union[AbstractExampleRetriever, Sequence[ExampleTuple], None] = None,
        instruction_template: PromptTemplate = DEFAULT_INSTRUCTION_PROMPT,
        input_key: str = "query",
        # Smirking kat language
        language_name: str = "😼",
        input_formatter: InputFormatter = "markdown_text",
    ) -> CodeChain:
        """Create a code chain from pre-configured defaults.

        Args:
            llm: The language model to use for coding
            interpreter: The interpreter that will be used to execute the program
            ast_printer: An ast printer that can print the AST as a text string
            context: Either a list of functions or a context retriever.
                               The list of functions can be a mixture of python
                               callables and Kork external functions.
                               All python functions will be converted into Kork
                               external functions, and everything passed into
                               the default context retriever.
            examples: An example retriever or a list of examples.
                       If a list of examples, a simple example retriever
                       will be created.
            instruction_template: Use to customize the instruction components of
                                  the prompt.
            language_name: The language name to use for the prompt.
            input_formatter: A formatting that will be applied to the input part
                             of each example tuple, if passing in a list of examples.
            input_key: The input key to use.

        Returns:
            A code chain
        """
        if isinstance(examples, (AbstractExampleRetriever, type(None))):
            _example_retriever = examples
        elif isinstance(examples, Sequence):
            formatted_examples = format_examples(
                language_name,
                examples,
                ast_printer,
                input_formatter=input_formatter,
            )
            _example_retriever = SimpleExampleRetriever(examples=formatted_examples)
        else:
            raise TypeError(
                f"example_retriever must be of type `AbstractExampleRetriever` or "
                f"`Sequence[ExampleTuple]`. Got {type(examples)} instead."
            )

        if isinstance(context, (AbstractContextRetriever, type(None))):
            _context_retriever = context
        elif isinstance(context, Sequence):
            _context_retriever = SimpleContextRetriever.from_functions(context)
        else:
            raise TypeError(
                f"context_retriever must be of type `AbstractContextRetriever` or "
                f"`Sequence[FunctionDefinition]`. Got {type(context)} "
                f"instead."
            )

        return cls(
            llm=llm,
            interpreter=interpreter,
            ast_printer=ast_printer,
            context_retriever=_context_retriever,
            example_retriever=_example_retriever,
            instruction_template=instruction_template,
            input_key=input_key,
            language_name=language_name,
            input_formatter=input_formatter,
        )



# type:ignore[no-untyped-def]
"""Kork's default AST parser.

Kork uses Lark to parse the AST. The grammar follows closely
the one used in Crafting Interpreters for the Lox Programming Language.

https://craftinginterpreters.com/appendix-i.html#expressions

The grammar and parser were clobbered together in a few hours of work.

If you deeply care about language design, please forgive any transgressions,
and feel free to help improve the grammar/parser/interpreter.
"""
from typing import cast

from lark import Lark, Transformer, v_args

from kork import ast

GRAMMAR = """
    program: statement+

    statement: function_decl
              | extern_function_decl
              | var_decl
              | expr

    function_decl: "fn" CNAME "(" [params] ")" "->" type block
    extern_function_decl: "extern" "fn" CNAME "(" [params] ")" "->" type
    var_decl: ("let" | "const" | "var") CNAME "=" expr [";"]
    params: param ("," param)*
    param: CNAME ":" type
    type: CNAME
    
    block: "{" [statement+] "}"
    
    ?expr: assignment
    
    ?assignment: CNAME "=" expr [";"] -> variable_assign
        | term
        
    !?term: term "+" factor -> binary
        | term "-" factor -> binary
        | factor
        
    !?factor: factor "*" unary -> binary
        | factor "/" unary -> binary
        | factor "%" unary -> binary
        | unary
        
    !?unary: "-" unary -> unary_r
        | "+" unary -> unary_r
        | power
        
    !?power: call "**" unary -> binary
        | call "^" unary -> binary
        | call
    
    ?call: CNAME "(" [args] ")" -> function_call
        | primary
    
    ?primary: numbers
        | list
        | string
        | "true" -> true
        | "false" -> false
        | "null" -> null
        | "(" expr ")" -> group
        | CNAME -> identifier
        
    ?numbers: SIGNED_NUMBER ("E"|"e") SIGNED_NUMBER -> scientific
        | SIGNED_NUMBER -> number
        
    args: expr ("," expr)*
    string: ESCAPED_STRING
    list: "[" [args] "]"

    %import common.CNAME
    %import common.SIGNED_NUMBER
    %import common.ESCAPED_STRING
    %import common.WS
    %ignore WS
    %ignore /#.*/
    %ignore /\/\/.*/
"""


# mypy: ignore-errors


@v_args(inline=True)
class AstParser(Transformer):
    """An ast parser for the given language."""

    def program(self, *items) -> ast.Program:
        """Create a program from the given statements."""
        return ast.Program(stmts=items)

    def function_decl(self, name, params, return_type, block) -> ast.FunctionDef:
        """Function declaration from the given name, params, return type and block."""
        return ast.FunctionDef(
            name=str(name),
            params=params,
            return_type=return_type,
            body=block,
        )

    def extern_function_decl(self, name, params, return_type) -> ast.ExternFunctionDef:
        """Extern function declaration from the given name, params and return type."""
        return ast.ExternFunctionDef(
            name=str(name),
            params=params or ast.ParamList(params=[]),
            return_type=return_type,
        )

    def statement(self, item) -> ast.Stmt:
        """Create a statement from the given item."""
        return item

    def var_decl(self, name, value) -> ast.VarDecl:
        """Variable declaration from the given name and value."""
        return ast.VarDecl(
            name=str(name),
            value=value,
        )

    def params(self, *items) -> ast.ParamList:
        """Create a param list from the given items."""
        return ast.ParamList(params=items or [])

    def unary_r(self, operator, right) -> ast.Unary:
        """Create a unary from the given operator and left."""
        op = str(operator)
        if op not in {"+", "-"}:
            raise ValueError(f"Unknown operator: {op}.")
        return ast.Unary(operator=str(operator), right=right)

    def binary(self, left, operator, right) -> ast.Binary:
        """Create a binary from the given left, operator and right."""
        op = str(operator)
        if op not in {"+", "-", "*", "/", "**", "//", "%", "^"}:
            raise ValueError(f"Unknown operator: {op}.")
        return ast.Binary(left=left, operator=str(operator), right=right)

    def param(self, name, type_) -> ast.Param:
        """Create a param from the given name and type."""
        return ast.Param(
            name=str(name),
            type_=type_,
        )

    def type(self, name):
        """Create a type from the given name."""
        return str(name)

    def block(self, *statements):
        """Create a block from the given statements."""
        return statements or []

    def group(self, expr) -> ast.Grouping:
        """Create a group from the given expression."""
        return ast.Grouping(expr=expr)

    def function_call(self, name, args) -> ast.FunctionCall:
        """Create a function call from the given name and args."""
        return ast.FunctionCall(name=str(name), args=args or [])

    def identifier(self, name) -> ast.Variable:
        """Create a variable reference from the given name."""
        return ast.Variable(
            name=str(name),
        )

    def variable_assign(self, name, value) -> ast.Assign:
        """Create a variable assignment from the given name and value."""
        return ast.Assign(
            name=str(name),
            value=value,
        )

    def expr(self, item) -> ast.Expr:
        """Create an expression from the given item."""
        return item

    def args(self, *items):
        """Create an argument list for a function from the given items."""
        return items or []

    def list(self, items) -> ast.List_:
        """Create a list literal from the given items."""
        return ast.List_(
            elements=items or [],
        )

    def true(self) -> ast.Literal:
        """Create a true literal."""
        return ast.Literal(value=True)

    def false(self) -> ast.Literal:
        """Create a false literal."""
        return ast.Literal(value=False)

    def null(self) -> ast.Literal:
        """Create a null literal."""
        return ast.Literal(value=None)

    def number(self, item) -> ast.Literal:
        try:
            return ast.Literal(value=int(item))
        except ValueError:
            return ast.Literal(value=float(item))

    def scientific(self, coefficient, exponent) -> ast.Literal:
        return ast.Literal(value=coefficient * 10**exponent)

    def string(self, item) -> ast.Literal:
        return ast.Literal(
            value=str(item[1:-1]),
        )


ast_parser = Lark(GRAMMAR, parser="lalr", transformer=AstParser(), start="program")

# PUBLIC API


def parse(source: str) -> ast.Program:
    """Parse the given source code into an AST."""
    return cast(ast.Program, ast_parser.parse(source))



import abc
from typing import Any, Union

from kork import ast


class AbstractAstPrinter(ast.Visitor, abc.ABC):
    @abc.abstractmethod
    def visit(
        self, element: Union[ast.Stmt, ast.Expr], pretty_print: bool = False
    ) -> str:
        """Entry-point for printing the AST."""


class AstPrinter(AbstractAstPrinter):
    """Default AST Printer implementation."""

    def visit(
        self, element: Union[ast.Stmt, ast.Expr], pretty_print: bool = False
    ) -> str:
        """Entry-point for printing the AST."""
        data = {
            "call_depth": 0,
            "pretty_print": pretty_print,
        }
        return element.accept(self, **data)

    def visit_program(self, program: ast.Program, **data: Any) -> str:
        """Print a program."""
        return "\n".join(stmt.accept(self, **data) for stmt in program.stmts)

    def visit_extern_function_def(
        self, extern_function_def: ast.ExternFunctionDef, **data: Any
    ) -> str:
        """Print an extern function definition."""
        code = [
            f"extern fn {extern_function_def.name}",
            "(",
            extern_function_def.params.accept(self, **data),
            ")",
            " -> ",
            extern_function_def.return_type,
        ]

        if extern_function_def.doc_string:
            # TODO: This is hacky for prototyping, will need to have a better
            # way to represent the doc string.
            doc_string = extern_function_def.doc_string.strip().split("\n")[0]
            code.append(f" // {doc_string}")

        return "".join(code)

    def visit_function_call(self, call: ast.FunctionCall, **data: Any) -> str:
        """Print a function call."""
        data["call_depth"] += 1
        if data["pretty_print"]:
            call_str = f"{call.name}(\n"
            call_str += ",\n".join(
                "    " * data["call_depth"] + arg.accept(self, **data)
                for arg in call.args
            )
            call_str += "\n" + "    " * (data["call_depth"] - 1) + ")"
        else:
            call_str = (
                f"{call.name}("
                f"{', '.join(arg.accept(self, **data) for arg in call.args)}"
                f")"
            )
        return call_str

    def visit_function_def(self, function_def: ast.FunctionDef, **data: Any) -> str:
        """Print a function definition."""
        signature = [
            f"fn {function_def.name}",
            "(",
            function_def.params.accept(self, **data),
            ")",
            " -> ",
            function_def.return_type,
        ]

        code = [" {"]
        for stmt in function_def.body:
            code.append(stmt.accept(self, **data))
        code.append("}")
        return "".join(signature) + "\n".join(code)

    def visit_variable(self, variable: ast.Variable, **data: Any) -> str:
        """Print a variable."""
        return variable.name

    def visit_var_decl(self, var_decl: ast.VarDecl, **data: Any) -> str:
        """Print a variable declaration."""
        return f"var {var_decl.name} = {var_decl.value.accept(self, **data)}"

    def visit_assign(self, assign: ast.Assign, **data: Any) -> str:
        """Print an assignment."""
        return f"{assign.name} = {assign.value.accept(self, **data)}"

    def visit_literal(self, literal: ast.Literal, **data: Any) -> str:
        """Print a literal."""
        value = literal.value
        if isinstance(value, type(None)):
            return "null"
        elif isinstance(value, bool):
            return str(value).lower()
        elif isinstance(value, str):
            return f'"{value}"'
        elif isinstance(value, (float, int)):
            return str(value)
        else:
            raise AssertionError(f"Unknown literal type: {type(value)}")

    def visit_param_list(self, param_list: ast.ParamList, **data: Any) -> str:
        """Print a parameter list."""
        return ", ".join(param.accept(self, **data) for param in param_list.params)

    def visit_param(self, param: ast.Param, **data: Any) -> str:
        """Print a parameter."""
        return f"{param.name}: {param.type_}"

    def visit_list_(self, list_: ast.List_, **data: Any) -> str:
        """Print a list."""
        return (
            f"[{', '.join(element.accept(self, **data) for element in list_.elements)}]"
        )

    def visit_unary(self, unary: ast.Unary, **data: Any) -> str:
        """Visit a unary expression."""
        return f"{unary.operator}{unary.right.accept(self, **data)}"

    def visit_grouping(self, grouping: ast.Grouping, **data: Any) -> str:
        return f"({grouping.expr.accept(self, **data)})"

    def visit_binary(self, binary: ast.Binary, **data: Any) -> str:
        """Print a binary expression."""
        return (
            f"{binary.left.accept(self, **data)} "
            f"{binary.operator} {binary.right.accept(self, **data)}"
        )



"""Utils for displaying chain results in a notebook."""
import base64
import math
from html import escape
from io import BytesIO
from typing import Any, Optional, Sequence, TypedDict, Union

from kork.ast_printer import AstPrinter
from kork.chain import CodeResult
from kork.parser import parse  # type: ignore

try:
    from PIL.Image import Image
except ImportError:
    Image = None


def image_base64(image: Image) -> str:
    """Get a base64 representation of an image."""
    with BytesIO() as buffer:
        image.save(buffer, "jpeg")
        return base64.b64encode(buffer.getvalue()).decode()


def as_img_tag(image: Image) -> str:
    """Get an HTML representation of an image."""
    return f'<img src="data:image/jpeg;base64,{image_base64(image)}">'


class HtmlResult(TypedDict):
    """A result that can be displayed in a notebook."""

    query: str
    errors: str
    raw: str
    code: str
    result: str
    expected: str
    correct: str


class _NoExpectedAnswer:
    """A sentinel class to indicate that there is no expected answer."""

    pass


NO_EXPECTED_ANSWER = _NoExpectedAnswer()


def as_html_dict(
    code_result: CodeResult,
    query: Optional[str] = None,
    expected_answer: Optional[Any] = None,
    result_key: str = "result",
    pretty_print: bool = True,
) -> HtmlResult:
    """Use to generate a dict that can be easily displayed in an IPython notebook."""
    from IPython import display as ipy_display

    code = code_result["code"].strip()

    if pretty_print:
        code = AstPrinter().visit(parse(code), pretty_print=True)
    else:
        code = code

    if code:
        _code = ipy_display.Code(data=code)._repr_html_().strip().replace("\n", "<br/>")
    else:
        _code = code

    env = code_result["environment"]

    result = env.variables.get(result_key, "") if env else ""

    if expected_answer is NO_EXPECTED_ANSWER:
        _correct = "N/A"
    else:
        if isinstance(expected_answer, (float, int)) and isinstance(
            result, (float, int)
        ):
            correct = math.isclose(expected_answer, result)
        else:
            correct = bool(expected_answer == result)
        _correct = "✅" if correct else "⛔"

    query = query or code_result.get("query", "")  # type: ignore

    if Image and isinstance(result, Image):
        _result = as_img_tag(result)
    else:
        _result = escape(str(result).strip())

    _expected = (
        "" if expected_answer is NO_EXPECTED_ANSWER else escape(str(expected_answer))
    )

    return {
        "query": escape(str(query)),
        "errors": escape(str(code_result["errors"])),
        "raw": escape(str(code_result["raw"])).replace("\n", "<br/>"),
        "code": _code,
        "result": _result,
        "expected": _expected,
        "correct": _correct,
    }


def display_html_results(
    html_results: Union[Sequence[HtmlResult], HtmlResult],
    columns: Optional[Sequence[str]] = None,
) -> Any:
    """Display a sequence of HTML results as a table."""
    import pandas as pd
    from IPython import display

    _results = html_results if isinstance(html_results, list) else [html_results]

    _columns = (
        columns
        if columns
        else ["query", "code", "result", "expected", "correct", "errors", "raw"]
    )

    df = pd.DataFrame(_results, columns=_columns)
    if "query" in df.columns:
        df["query"] = df["query"].str.wrap(40)
        df["query"] = df["query"].str.replace("\n", "<br/>")
    df = df.style.set_properties(**{"text-align": "left"})
    return display.HTML(df.to_html(escape=False, index=False))


def display_code_result(
    code_result: CodeResult,
    query: Optional[str] = None,
    expected_answer: Optional[Any] = NO_EXPECTED_ANSWER,
    columns: Optional[Sequence[str]] = None,
    result_key: str = "result",
) -> Any:
    """Display a code result as a table."""
    return display_html_results(
        as_html_dict(
            code_result,
            query=query,
            expected_answer=expected_answer,
            result_key=result_key,
        ),
        columns=columns,
    )



"""Logic that attempts to surface the most relevant information for writing code."""
from __future__ import annotations

import abc
import dataclasses
from typing import Callable, Sequence, Union

from kork import ast
from kork.foreign_funcs import to_extern_func_def


@dataclasses.dataclass(frozen=True)
class AbstractContextRetriever(abc.ABC):
    """Abstract interface for retrieving programming context."""

    @abc.abstractmethod
    def retrieve(self, query: str) -> Sequence[ast.ExternFunctionDef]:
        """Retrieve the external function definitions."""


FuncLike = Union[ast.ExternFunctionDef, Callable]


@dataclasses.dataclass(frozen=True)
class SimpleContextRetriever(AbstractContextRetriever):
    """Retrieve information as was provided without any filtering or re-ranking."""

    external_functions: Sequence[ast.ExternFunctionDef] = tuple()

    def __post_init__(self) -> None:
        """Validate the external functions."""
        for func in self.external_functions:
            if not isinstance(func, ast.ExternFunctionDef):
                raise ValueError(
                    "SimpleRetriever only accepts ExternFunctionDef objects in "
                    f"the external_functions list. Got: {func}"
                )

    def retrieve(self, query: str) -> Sequence[ast.ExternFunctionDef]:
        """Retrieve the external function definitions."""
        return self.external_functions

    @classmethod
    def from_functions(cls, mixed_funcs: Sequence[FuncLike]) -> SimpleContextRetriever:
        """Create a simple retrieval from a sequence of functions.

        Args:
            mixed_funcs: A sequence of functions or external function definitions.

        Returns:
            A simple retriever.
        """

        external_functions = []

        for func in mixed_funcs:
            if isinstance(func, ast.ExternFunctionDef):
                external_functions.append(func)
            else:
                external_functions.append(to_extern_func_def(func))

        return cls(external_functions=external_functions)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kork']

package_data = \
{'': ['*']}

install_requires = \
['langchain>=0.0.110',
 'lark>=1.1.5,<2.0.0',
 'openai>=0.27,<0.28',
 'sphinx-design>=0.4.1,<0.5.0']

setup_kwargs = {
    'name': 'kork',
    'version': '0.0.2',
    'description': 'Programming with LLMs',
    'long_description': '[![Unit Tests](https://github.com/langchain-ai/kork/actions/workflows/test.yml/badge.svg?branch=main&event=push)](https://github.com/langchain-ai/kork/actions/workflows/test.yml)\n\n# Kork ![alt The Parrot](assets/parrot.png)\n\n`Kork` is an *experimental* [Langchain chain](https://python.langchain.com/en/latest/modules/chains.html) that helps build natural language APIs powered by LLMs.\n\n## Features\n\n1. Assemble a natural language API from a set of python functions.\n2. Generate a prompt to help the LLM write a **correct** program.\n3. Execute the program generated by the LLM **safely**.\n4. Program generation takes a single LLM call, as opposed to agents that use a call per action taken.\n\n## How\n\nThe chain takes a user query, translates it into a program, and executes it using the `Kork` interpreter.\n\nThe interpreter is used to run code written in the `Kork` programming language.\n\nThe language is limited by design to variable declarations, function invocations and arithmetic operations.\n\nThis limitation allows controlling the kind of programs the LLMs can generate, making the output of the LLM more predictable and safer to execute in a production setting. (An agent with access to "constrained" tools has similar benefits, but requires a call per action taken.)\n\n`Kork` adds a few additional tricks (e.g., retriever interfaces for examples and foreign functions) to help guide the LLM to generate a **correct** program.\n\n## What?\n\n*No loops, no conditionals, no file access, no network access, no arbitrary code execution. WHAT?!*\n\nThe ability to invoke custom functions goes a long way in terms of the kinds of programs that can be written! (You can always add a `write_to_file` function!)\n\n\n## Limitations\n\n- `Kork` **cannot write arbitrary** code. If that\'s what you need, save yourself some time and use [docker](https://www.docker.com/) and a real programming language.\n- The `Kork` chain is **not** an Agent, so it cannot inspect intermediate steps, instead it generates a pre-defined set of function calls that need to be executed. (But remember that function calls can involve calls to an agent!)\n- The `Kork` language and interpreter are limited to *function invocation*, *variable declaration* and\n  *basic arithmetic* (no function declaration, loops etc.)\n- Only supporting `int`, `float`, `str`, `type(None)`, `bool` types. No support for `lists` or `object` types.\n- Very limited type annotations.\n\n## Quality\n\nThe short version: No benchmarks yet!\n\nThe long version: Quality of generated programs depends on many factors.\n\nYou can experiment with tweaking the prompt (e.g., explain the syntax of the language), changing the foreign function retriever (e.g., retrieve the most relevant foreign functions based on the user query), or providing examples in the form of (query, expected program).\n\nThe prompt, examples and syntax of the language can trip up the LLM to assume\nit\'s programming in a specific language (e.g., typescript or python) and assume\nthat it can use language features that are not supported by `Kork` or to import\nlibraries that are not available.\n\n## Safety\n\nNothing is bulletproof when humans are involved.\n\n* Don\'t do silly things like exposing `eval` as a foreign function.\n* If a foreign function allocates memory, the LLM could ask for more memory than is available crashing the process.\n* Think carefully before using the output of an LLM. If the LLM behaves maliciously, is it safe to use an LLM in the given context? \n\n## Future Work\n\nLet us know if you\'re interested in contributing or have ideas for improvements!\n\n- Allow enabling/disabling language features\n- Allow changing underlying language syntax\n- Add support for objects\n- Add other language features (e.g., loops) \n- Provide foreign function retriever implementations based on similarity to user query\n\n## Compatibility\n\n`Kork` is tested against python 3.8, 3.9, 3.10, 3.11.\n\n## Installation \n\n```sh\npip install kork\n```\n\n## 🙏 Appreciation\n\n* [Lark](https://github.com/lark-parser/lark) -- For making it easy to define a grammar and parse it!\n* [Bob Nystrom](https://github.com/munificent) -- For writing [Crafting Interpreters](https://www.craftinginterpreters.com/)!\n\n## © Contributing\n\nIf you have any ideas or feature requests, please open an issue and share!\n\nSee [CONTRIBUTING.md](https://github.com/langchain-ai/kork/blob/main/CONTRIBUTING.md) for more information.\n\n## 🎶 Why the name?\n\nFast to type and maybe sufficiently unique.\n',
    'author': 'LangChain',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/langchain-ai/kork',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.1,<4.0.0',
}


setup(**setup_kwargs)



from typing import Any, Optional, Sequence, TypedDict, Union

from lark.exceptions import LarkError

from kork import ast
from kork.environment import Environment
from kork.exceptions import KorkRunTimeException

# TODO: Determine why mypy is not recognizing the import.
from kork.parser import parse  # type: ignore[attr-defined]


class Interpreter(ast.Visitor):
    """Kork's default interpreter."""

    def __init__(self, program: ast.Program, environment: Environment) -> None:
        """Initialize the interpreter."""
        self.program = program
        self.environment = environment
        self.last_expression_value: Any = None

    def intepret(self) -> None:
        """Interpret the program."""
        self.program.accept(self)

    def visit_program(self, program: ast.Program) -> None:
        """Visit a program."""
        for stmt in program.stmts:
            stmt.accept(self)

    def visit_extern_function_def(
        self, extern_function_def: ast.ExternFunctionDef
    ) -> None:
        """Visit an external function definition."""
        self.environment.variables[extern_function_def.name] = extern_function_def

    def visit_function_call(self, call: ast.FunctionCall) -> Any:
        """Visit a function call."""
        function = self.environment.get_symbol(call.name)
        args = [arg.accept(self) for arg in call.args]
        if isinstance(function, ast.ExternFunctionDef):
            if function.implementation is None:
                raise KorkRunTimeException("External function has not been linked yet.")
            try:
                return function.implementation(*args)
            except Exception as e:
                raise KorkRunTimeException(
                    "Encountered an exception while invoking an external function "
                    f"`{call.name}` with arguments: `{args}`."
                    f"Details: {e}"
                )
        elif isinstance(function, ast.FunctionDef):
            raise KorkRunTimeException("Function definition is unimplemented.")
        else:
            raise KorkRunTimeException(f"Cannot call {call.name}.")

    def visit_function_def(self, function_def: ast.FunctionDef) -> None:
        """Visit a function definition."""
        self.environment.variables[function_def.name] = function_def

    def visit_variable(self, variable: ast.Variable) -> Any:
        """Visit a variable."""
        return self.environment.get_symbol(variable.name)

    def visit_var_decl(self, var_decl: ast.VarDecl) -> None:
        """Visit a variable declaration."""
        self.environment.variables[var_decl.name] = var_decl.value.accept(self)

    def visit_unary(self, unary: ast.Unary) -> Any:
        """Visit a unary expression."""
        operator = unary.operator
        right = unary.right.accept(self)
        if operator == "-":
            return -right
        elif operator == "+":
            return right
        else:
            raise ValueError(f"Unknown operator {operator}.")

    def visit_binary(self, binary: ast.Binary) -> Any:
        """Visit a binary expression."""
        left = binary.left.accept(self)
        right = binary.right.accept(self)
        operator = binary.operator
        if operator == "+":
            return left + right
        elif operator == "-":
            return left - right
        elif operator == "*":
            return left * right
        elif operator == "/":
            return left / right
        elif operator == "**":
            return left**right
        elif operator == "%":
            return left % right
        elif (
            operator == "^"
        ):  # Please note that we do not follow python convention here
            return left**right  # Duplicating `**` operator
        else:
            raise ValueError(f"Unknown operator {operator}.")

    def visit_grouping(self, grouping: ast.Grouping) -> Any:
        """Visit a grouping."""
        return grouping.expr.accept(self)

    def visit_assign(self, assign: ast.Assign) -> None:
        """Visit an assignment."""
        self.environment.variables[assign.name] = assign.value.accept(self)

    def visit_literal(self, literal: ast.Literal) -> Union[int, float, bool, str, None]:
        """Visit a number."""
        return literal.value

    def visit_list_(self, list_: ast.List_) -> list:
        """Visit a list."""
        return [element.accept(self) for element in list_.elements]


# PUBLIC API


class InterpreterResult(TypedDict):
    """Use this to return the result of the interpreter."""

    environment: Environment
    errors: Sequence[Exception]


def run_interpreter(
    code: str, environment: Optional[Environment] = None
) -> InterpreterResult:
    """Run the interpreter with the given code.

    Args:
        code: The code to run
        environment: The environment to run the code in

    Returns:
        the final environment after running the code (this will likely change)
    """
    # TODO(Eugene): May want to refactor to avoid cloning here without introducing
    #                mutability associated bugs.
    environment = (
        environment.clone() if environment else Environment(parent=None, variables={})
    )
    try:
        program = parse(code)
        interpreter = Interpreter(program, environment)
        interpreter.intepret()
    except (
        LarkError,
        KorkRunTimeException,
        ValueError,
        ZeroDivisionError,
        TypeError,
    ) as e:
        return InterpreterResult(
            environment=environment,
            errors=[e],
        )

    return InterpreterResult(
        environment=interpreter.environment,
        errors=[],
    )
